==> fxusb-suse7.1-03.09.10.tar.gz: fritz/src.drv/tools.c <==
/* 
 * tools.c
 * Copyright (C) 2002, AVM GmbH. All rights reserved.
 * 
 * This Software is  free software. You can redistribute and/or
 * modify such free software under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * The free software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this Software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA, or see
 * http://www.opensource.org/licenses/lgpl-license.html
 * 
 * Contact: AVM GmbH, Alt-Moabit 95, 10559 Berlin, Germany, email: info@avm.de
 */

#include <asm/atomic.h>
#include <asm/system.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/vmalloc.h>
#include <linux/kernel.h>
#include "defs.h"
#include "tools.h"

typedef struct __header {
    unsigned		type;
#if !defined (NDEBUG)
    unsigned		size;
    unsigned		from;
    unsigned		tag;
#endif
} header_t;

#define	TYPE_NONE	'?'
#define TYPE_KMALLOCED	'k'
#define	TYPE_VMALLOCED	'v'
#define	PRIORITY	GFP_ATOMIC
#define KMALLOC_LIMIT	131072

#if !defined (NDEBUG)
#include <asm/atomic.h>

#define	FENCE_TAG	0xDEADBEEF

static atomic_t		alloc_count = ATOMIC_INIT (0);

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
unsigned hallocated (void) {
    
	return atomic_read (&alloc_count);
} /* hallocated */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
unsigned hallocator (void * mem) {
	header_t * hdr;

	if (mem != NULL) {
		hdr = ((header_t *) mem) - 1;
		return (hdr->tag != FENCE_TAG) ? 0 : hdr->from;
	} else {
		return 0;
	}
} /* hallocator */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
int hvalid (void * mem) {
	header_t *	hdr;
	int		flag = TRUE;

	if (mem != NULL) {
		hdr  = ((header_t *) mem) - 1;
		flag = (hdr->tag == FENCE_TAG) && 
		   (* (unsigned *) (((char *) mem) + hdr->size) == FENCE_TAG);
	} 
	return flag;
} /* hvalid */
#endif

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void * ALLOCATE (unsigned n, unsigned * t) {
	void *		temp;
	unsigned	type;

	if (n <= KMALLOC_LIMIT) {
		temp = kmalloc (n, PRIORITY);
		type = TYPE_KMALLOCED;
	} else {
		temp = vmalloc (n);
		type = TYPE_VMALLOCED;
	}
	*t = (temp != NULL) ? type : TYPE_NONE;
	return temp;
} /* ALLOCATE */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
#if !defined (NDEBUG)
#define	PATCH(n)	sizeof(header_t)+sizeof(unsigned)+((n)?(n):1)
#else
#define	PATCH(n)	sizeof(header_t)+((n)?(n):1)
#endif

static void * halloc (unsigned size, unsigned addr) {
	unsigned	n, t = TYPE_NONE;
	void *		mem;
	header_t *	hdr;

	n = PATCH(size);
	if (NULL == (hdr = (header_t *) ALLOCATE (n, &t))) {
		log ("Memory request (%u/%u bytes) failed.\n", 
							size, PATCH(size));
		mem = NULL;
	} else {
		mem = (void *) (hdr + 1);
		hdr->type = t;
#if !defined (NDEBUG)
#if defined (LOG_MEM_BLOCKS)
		printk (KERN_INFO "ALLOC %6u bytes @ %08X\n", size, addr);
#endif
		hdr->size = size ? size : 1;
		hdr->from = addr;
		hdr->tag  = FENCE_TAG;
		* (unsigned *) (((char *) mem) + size) = FENCE_TAG;
		atomic_add (size, &alloc_count);
#else
		UNUSED_ARG (addr);
#endif
	}
	return mem;
} /* halloc */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void * hmalloc (unsigned size) {

	return halloc (size, *(((unsigned *) &size) - 1));
} /* hmalloc */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void * hcalloc (unsigned size) {
	void * mem;

	mem = halloc (size, *(((unsigned *) &size) - 1));
	if ((mem != NULL) && (size != 0)) {
		memset (mem, 0, size);
	}
	return mem;
} /* hcalloc */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void hfree (void * mem) {
	header_t *	hdr;
#if !defined (NDEBUG)
	int		fence1, fence2;
#endif

	if (mem != NULL) {
		hdr = ((header_t *) mem) - 1;
#if !defined (NDEBUG)
#if defined (LOG_MEM_BLOCKS)
		printk (KERN_INFO "FREE  %6u bytes @ %08X\n", hdr->size, hdr->from);
#endif
		fence1 = (hdr->tag == FENCE_TAG);
		fence2 = (* (unsigned *) (((char *) mem) + hdr->size) 
								== FENCE_TAG);
		if (!(fence1 && fence2)) {
			log ("FENCE VIOLATED (%u/0x%08X)!\n", 
							hdr->size, hdr->from);
		}
		atomic_sub (hdr->size, &alloc_count);
#endif
		assert ((hdr->type == TYPE_KMALLOCED) || 
						(hdr->type == TYPE_VMALLOCED));
		if (TYPE_KMALLOCED == hdr->type) {
			kfree (hdr);
		} else {
			vfree (hdr);
		}
	}
} /* hfree */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void vlprintf (const char * level, const char * fmt, va_list args) {
	static char	line[256];

	vsprintf (line, fmt, args);
	printk ("%s%s: %s", level, TARGET, line);
} /* vlprintf */
 
/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void lprintf (const char * level, const char * fmt, ...) {
	va_list args;

	va_start (args, fmt);
	vlprintf (level, fmt, args);
	va_end (args);
} /* lprintf */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void message (const char * fmt, ...) {
	va_list args;

	va_start (args, fmt);
	vlprintf (KERN_INFO, fmt, args);
	va_end (args);
} /* message */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
#ifndef NDEBUG
void _OSassert (void * exp, void * file, unsigned line) {

	message ("assert (%s) in %s(%u)\n", exp, file, line);
} /* _OSassert */

void _OSinfo (void * exp, void * file, unsigned line) {

	message ("info (%s) in %s(%u)\n", exp, file, line);
} /* _OSinfo */
#endif

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

unsigned long atomic_xchg (volatile atomic_t * v, unsigned value) {

	return __xchg (value, &v->counter, sizeof (unsigned));
} /* atomic_xchg */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

==> fxusb-suse8.1-03.10.02.tar.gz: fritz/src.drv/tools.c <==
/* 
 * tools.c
 * Copyright (C) 2002, AVM GmbH. All rights reserved.
 * 
 * This Software is  free software. You can redistribute and/or
 * modify such free software under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * The free software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this Software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA, or see
 * http://www.opensource.org/licenses/lgpl-license.html
 * 
 * Contact: AVM GmbH, Alt-Moabit 95, 10559 Berlin, Germany, email: info@avm.de
 */

#include <asm/atomic.h>
#include <asm/system.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/vmalloc.h>
#include <linux/kernel.h>
#include "defs.h"
#include "tools.h"

typedef struct __header {
    unsigned		type;
#if !defined (NDEBUG)
    unsigned		size;
    unsigned		from;
    unsigned		tag;
#endif
} header_t;

#define	TYPE_NONE	'?'
#define TYPE_KMALLOCED	'k'
#define	TYPE_VMALLOCED	'v'
#define	PRIORITY	GFP_ATOMIC
#define KMALLOC_LIMIT	131072

#if !defined (NDEBUG)
#include <asm/atomic.h>

#define	FENCE_TAG	0xDEADBEEF

static atomic_t		alloc_count = ATOMIC_INIT (0);

#if !defined (NDEBUG) && defined (LOG_TIMER)
static struct timeval	zero_time;
#endif

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
unsigned hallocated (void) {
    
	return atomic_read (&alloc_count);
} /* hallocated */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
unsigned hallocator (void * mem) {
	header_t * hdr;

	if (mem != NULL) {
		hdr = ((header_t *) mem) - 1;
		return (hdr->tag != FENCE_TAG) ? 0 : hdr->from;
	} else {
		return 0;
	}
} /* hallocator */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
int hvalid (void * mem) {
	header_t *	hdr;
	int		flag = TRUE;

	if (mem != NULL) {
		hdr  = ((header_t *) mem) - 1;
		flag = (hdr->tag == FENCE_TAG) && 
		   (* (unsigned *) (((char *) mem) + hdr->size) == FENCE_TAG);
	} 
	return flag;
} /* hvalid */
#endif

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void * ALLOCATE (unsigned n, unsigned * t) {
	void *		temp;
	unsigned	type;

	if (n <= KMALLOC_LIMIT) {
		temp = kmalloc (n, PRIORITY);
		type = TYPE_KMALLOCED;
	} else {
		temp = vmalloc (n);
		type = TYPE_VMALLOCED;
	}
	*t = (temp != NULL) ? type : TYPE_NONE;
	return temp;
} /* ALLOCATE */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
#if !defined (NDEBUG)
#define	PATCH(n)	sizeof(header_t)+sizeof(unsigned)+((n)?(n):1)
#else
#define	PATCH(n)	sizeof(header_t)+((n)?(n):1)
#endif

static void * halloc (unsigned size, unsigned addr) {
	unsigned	n, t = TYPE_NONE;
	void *		mem;
	header_t *	hdr;

	n = PATCH(size);
	if (NULL == (hdr = (header_t *) ALLOCATE (n, &t))) {
		log ("Memory request (%u/%u bytes) failed.\n", 
							size, PATCH(size));
		mem = NULL;
	} else {
		mem = (void *) (hdr + 1);
		hdr->type = t;
#if !defined (NDEBUG)
#if defined (LOG_MEM_BLOCKS)
		printk (KERN_INFO "ALLOC %6u bytes @ %08X\n", size, addr);
#endif
		hdr->size = size ? size : 1;
		hdr->from = addr;
		hdr->tag  = FENCE_TAG;
		* (unsigned *) (((char *) mem) + size) = FENCE_TAG;
		atomic_add (size, &alloc_count);
#else
		UNUSED_ARG (addr);
#endif
	}
	return mem;
} /* halloc */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void * hmalloc (unsigned size) {

	return halloc (size, *(((unsigned *) &size) - 1));
} /* hmalloc */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void * hcalloc (unsigned size) {
	void * mem;

	mem = halloc (size, *(((unsigned *) &size) - 1));
	if ((mem != NULL) && (size != 0)) {
		memset (mem, 0, size);
	}
	return mem;
} /* hcalloc */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void hfree (void * mem) {
	header_t *	hdr;
#if !defined (NDEBUG)
	int		fence1, fence2;
#endif

	if (mem != NULL) {
		hdr = ((header_t *) mem) - 1;
#if !defined (NDEBUG)
#if defined (LOG_MEM_BLOCKS)
		printk (KERN_INFO "FREE  %6u bytes @ %08X\n", hdr->size, hdr->from);
#endif
		fence1 = (hdr->tag == FENCE_TAG);
		fence2 = (* (unsigned *) (((char *) mem) + hdr->size) 
								== FENCE_TAG);
		if (!(fence1 && fence2)) {
			log ("FENCE VIOLATED (%u/0x%08X)!\n", 
							hdr->size, hdr->from);
		}
		atomic_sub (hdr->size, &alloc_count);
#endif
		assert ((hdr->type == TYPE_KMALLOCED) || 
						(hdr->type == TYPE_VMALLOCED));
		if (TYPE_KMALLOCED == hdr->type) {
			kfree (hdr);
		} else {
			vfree (hdr);
		}
	}
} /* hfree */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

#if !defined (NDEBUG) && defined (LOG_TIMER)

#include <linux/time.h>

void setup_timer (dbg_timer * t, long dsec, long dusec) {
	
	assert (t != NULL);
	memset (&t->t, 0, sizeof (t->t));
	t->d.tv_sec  = dsec;
	t->d.tv_usec = dusec;
} /* dbg_setup_timer */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
int check_timer (dbg_timer * t) {
	int		res = 1;
	struct timeval	now;
	struct timeval	delta;
	
	assert (t != NULL);
	do_gettimeofday (&now);
	timeval_less (now, zero_time, &delta);
	now = delta;
	timeval_less (now, t->t, &delta);
	if ((delta.tv_sec > t->d.tv_sec) 
	|| ((delta.tv_sec == t->d.tv_sec) && (delta.tv_usec > t->d.tv_usec))
	) {
		lprintf (
			KERN_INFO, 
			"Timer '%s' exceeded: %ld s, %ld s\n", 
			t->name,
			delta.tv_sec,
			delta.tv_usec
		);
		res = 0;
	} 
	return res;
} /* check_timer */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void touch_timer (dbg_timer * t) {
	struct timeval	temp, delta;
	
	assert (t != NULL);
	do_gettimeofday (&temp);
	timeval_less (temp, zero_time, &delta);
	t->t = delta;
} /* touch_timer */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void start_watch (dbg_timer * w) {
	struct timeval	temp, delta;
	
	assert (w != NULL);
	do_gettimeofday (&temp);
	timeval_less (temp, zero_time, &delta);
	w->t = delta;
} /* start_watch */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void stop_watch (dbg_timer * w) {
	struct timeval	temp, delta;
	
	assert (w != NULL);
	do_gettimeofday (&temp);
	timeval_less (temp, zero_time, &delta);
	temp = delta;
	timeval_less (temp, w->t, &delta);
	w->t = delta;
} /* stop_watch */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
int timers_start (void) {

	do_gettimeofday (&zero_time);
	return 1;
} /* timers_start */
     
/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void timers_stop (void) {

} /* timers_stop */

#endif /* !NDEBUG && LOG_TIMER */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void vlprintf (const char * level, const char * fmt, va_list args) {
	static char	line[1024];

	VSNPRINTF (line, sizeof (line), fmt, args);
	printk ("%s%s: %s", level, TARGET, line);
} /* vlprintf */
 
/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void lprintf (const char * level, const char * fmt, ...) {
	va_list args;

	va_start (args, fmt);
	vlprintf (level, fmt, args);
	va_end (args);
} /* lprintf */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void message (const char * fmt, ...) {
	va_list args;

	va_start (args, fmt);
	vlprintf (KERN_INFO, fmt, args);
	va_end (args);
} /* message */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
#ifndef NDEBUG
void _OSassert (void * exp, void * file, unsigned line) {

	message ("assert (%s) in %s(%u)\n", exp, file, line);
} /* _OSassert */

void _OSinfo (void * exp, void * file, unsigned line) {

	message ("info (%s) in %s(%u)\n", exp, file, line);
} /* _OSinfo */
#endif

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

unsigned long atomic_xchg (volatile atomic_t * v, unsigned value) {

	return __xchg (value, &v->counter, sizeof (unsigned));
} /* atomic_xchg */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
